import funcy as fn

from discrete_signals import signal


DATA1 = [(0, 1), (1, 1.1), (2, 3)]


def test_make_signal():
    sig1 = signal(DATA1, start=0, end=4, tag='x')
    sig2 = signal(DATA1, start=1, end=2, tag='x')
    assert sig1 | sig2 == sig1

    assert set(fn.pluck(0, DATA1)) == set(sig1.times())
    assert set(fn.pluck(0, DATA1)) > set(sig2.times())
    assert len(sig2.times()) == 1
    assert set(fn.pluck(1, DATA1)) == {v['x'] for v in sig1.values()}


def test_repr():
    out = repr(signal(DATA1, start=0, end=4, tag='x'))
    assert out == "start, end: [0, 4)\n" \
        "data: [(0, {'x': 1}), (1, {'x': 1.1}), (2, {'x': 3})]"


def test_slice():
    sig1 = signal(DATA1, start=0, end=4, tag='x')
    sig2 = sig1[-float('inf'):float('inf')][0:4]
    assert sig1 == sig2

    assert sig1[1:2] == signal(DATA1, start=1, end=2, tag='x')
    assert len(sig1[1:2].values()) == 1


def test_time_shifts():
    sig1 = signal(DATA1, start=0, end=4, tag='x')
    assert (sig1 << 2) == (sig1 >> -2)
    assert (sig1 << 2) >> 2 == sig1


def test_append():
    sig1 = signal(DATA1, start=0, end=4, tag='x')
    sig2 = sig1 @ sig1
    assert len(sig2.values()) == 2*len(sig1.values())
    assert sig1.start == sig2.start
    assert sig2.end == 2*sig1.end


def test_par_compose():
    sig1 = signal(DATA1, start=0, end=4, tag='x')
    assert sig1 | sig1 == sig1

    sig2 = sig1 | sig1 >> 0.5
    assert len(sig2.values()) == 2*len(sig1.values())

    sig3 = sig1 | sig1 >> 1
    assert len(sig3.values()) == len(sig1.values())+1

    sig4 = sig1.map(lambda v: -v['x'], tag='y')
    sig5 = (sig1 | sig4).map(lambda v: sum(v.values()), tag='z')
    assert len(sig5.times()) == len(sig1.times())
    assert all(v['z'] == 0 for v in sig5.values())


def test_retag():
    sig1 = signal(DATA1, start=0, end=4, tag='x')
    assert sig1 == sig1.retag({'w': 'n'})

    sig2 = sig1.retag({'x': 'y'})
    assert sig2.tags == {'y'}

    assert sig2.retag({'y': 'x'}) == sig1


def test_rolling():
    sig1 = signal(DATA1, start=0, end=4, tag='x')
    sig2 = sig1.rolling(0, 3)
    assert {v['x'] for v in sig1.values()} == set(sig2[0]['x'])

    sig3 = sig1.rolling(0, 2)
    assert {v['x'] for v in sig1[:2].values()} == set(sig3[0]['x'])

    sig4 = sig1.rolling(0, 1)
    assert {v['x'] for v in sig1[:1].values()} == set(sig4[0]['x'])

    sig5 = sig1.rolling(-1.1, 1.1)
    assert {v['x'] for v in sig1.values()} == set(sig5[1.1]['x'])
    assert {v['x'] for v in sig1[2:].values()} == set(sig5[3.1]['x'])


def test_interp():
    sig1 = signal(DATA1, start=0, end=4, tag='x')
    for i in range(3):
        assert sig1[0] == sig1.interp(i/3)



from setuptools import find_packages, setup

DESC = 'A domain specific language for modeling and manipulating discrete time signals.'  # noqa

setup(
    name='discrete-signals',
    version='0.4.0',
    description=DESC,
    url='http://github.com/mvcisback/DiscreteSignals',
    author='Marcell Vazquez-Chanlatte',
    author_email='marcell.vc@eecs.berkeley.edu',
    license='MIT',
    install_requires=[
        'attrs',
        'funcy',
        'sortedcontainers',
    ],
    packages=find_packages(),
)



from collections import defaultdict
from typing import DefaultDict, Mapping, TypeVar

import attr
import funcy as fn
from sortedcontainers import SortedDict


Time, Value = TypeVar('Time'), TypeVar('Value')


@attr.s(slots=True, repr=False)
class DiscreteSignal:
    data = attr.ib(
        converter=SortedDict,
        type=Mapping[Time, DefaultDict[str, Value]]
    )
    start = attr.ib(type=Time)
    end = attr.ib(type=Time)

    def values(self):
        return self.data.values()

    def times(self):
        return self.data.keys()

    def items(self):
        return self.data.items()

    @property
    def tags(self):
        return set(fn.cat(self.values()))

    def evolve(self, **kwargs):
        return attr.evolve(self, **kwargs)

    def __repr__(self):
        return f"start, end: [{self.start}, {self.end})\n" \
            f"data: {[(t, dict(val)) for t, val in self.items()]}"

    def __rshift__(self, delta):
        return self.evolve(
            data=fn.walk_keys(lambda t: t + delta, self.data),
            start=self.start + delta,
            end=self.end + delta,
        )

    def __lshift__(self, delta):
        return self >> -delta

    def __matmul__(self, other):
        return self.evolve(
            data=fn.merge(
                self.data,
                fn.walk_keys(lambda t: t + self.end, other.data)
            ),
            end=self.end + (other.end - other.start)
        )

    def __or__(self, other):
        return self.evolve(
            data=fn.merge_with(lambda x: fn.merge(*x), self.data, other.data),
            start=min(self.start, other.start),
            end=max(self.end, other.end),
        )

    def __getitem__(self, key):
        if isinstance(key, slice):
            assert key.step is None
            start = self.start if key.start is None else key.start
            end = self.end if key.stop is None else key.stop

            return self.evolve(
                data=fn.select_keys(lambda t: start <= t < end, self.data),
                start=start, end=end,
            )
        return self.data[key]

    def rolling(self, start, end):
        if start != 0:
            return self.rolling(0, end-start) << start

        def apply_window(time_val):
            t, _ = time_val
            values = self[start + t: end + t].values()
            # Note: {} forces application of tuple.
            values = fn.merge_with(tuple, {}, *values)
            return (t, values)

        return self.evolve(
            data=fn.walk(apply_window, self.data),
            end=self.end - end
        )

    def map(self, func, tag=None):
        data = fn.walk_values(func, list(self.items()))
        return signal(data, self.start, self.end, tag)

    def retag(self, mapping):
        def _retag(val):
            return fn.walk_keys(lambda k: mapping.get(k, k), val)

        return self.evolve(data=fn.walk_values(_retag, self.data))

    def interp(self, t):
        assert self.start <= t < self.end
        key = self.data.iloc[self.data.bisect_right(t) - 1]
        return self[key]


def signal(data, start, end, tag=None):
    data = map(lambda x: (x[0], defaultdict(None, {tag: x[1]})), data)
    return DiscreteSignal(data=data, start=start, end=end)[start:end]
